/*
 * Copyright 2006 Takahiro Nakamura.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package woolpack.web;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import junit.framework.TestCase;
import woolpack.fn.Fn;
import woolpack.utils.Utils;

public class WebUtilsTest extends TestCase {

	public void testClearSession() {
		final SimpleWebContext c = new SimpleWebContext();
		c.setSession(Utils.concurrentMap(
				new HashMap<String, Object>(), new Object()));
		c.getSession().put("key0", "value0");
		c.getSession().put("key1", "value1");
		assertFalse(c.getSession().isEmpty());
		WebUtils.CLEAR_SESSION.exec(c);
		assertTrue(c.getSession().isEmpty());
	}
	
	public void testAccessor() throws Exception {
		final SimpleWebContext c = new SimpleWebContext();
		c.setInput(new HashMap<String, List<Object>>());
		c.setRequest(new HashMap<String, Object>());
		c.setSession(Utils.concurrentMap(new HashMap<String, Object>(), new Object()));
		c.setApplication(Utils.concurrentMap(new HashMap<String, Object>(), new Object()));
		c.setContainer(new HashMap());

		assertSame(c.getInput(), WebUtils.INPUT_EL.getValue(c));
		assertSame(c.getRequest(), WebUtils.REQUEST_EL.getValue(c));
		assertSame(c.getSession(), WebUtils.SESSION_EL.getValue(c));
		assertSame(c.getApplication(), WebUtils.APPLICATION_EL.getValue(c));
		assertSame(c.getContainer(), WebUtils.CONTAINER_EL.getValue(c));

		assertSame(c.getInput(), WebUtils.inputFn().exec(c));
		assertSame(c.getRequest(), WebUtils.requestFn().exec(c));
		assertSame(c.getSession(), WebUtils.sessionFn().exec(c));
		assertSame(c.getApplication(), WebUtils.applicationFn().exec(c));
		assertSame(c.getContainer(), WebUtils.containerFn().exec(c));
	}

	public void testConvertNormal() {
		assertEquals(Utils
				.map("key0", Arrays.asList("value00", "value01"))
				.map("key1", Arrays.asList("value10", "value11", "value12")),
				WebUtils.convert(Utils
						.map("key0", new String[] { "value00", "value01" })
						.map("key1", new String[] { "value10", "value11", "value12" })));
	}

	public void testConvertKeyClassCast() {
		final Map<Integer, String[]> base = Utils
			.map(Integer.valueOf(1), new String[] { "value00", "value01" });
		try {
			WebUtils.convert(base);
			fail();
		} catch (final ClassCastException expected) {
		}
	}
	
	public void testCheckTransactionToken() {
		final Fn<WebContext, Void, RuntimeException> fn = WebUtils.checkTransactionToken(
			"transactionToken",
			new Fn<WebContext, Boolean, RuntimeException>() {
				public Boolean exec(WebContext c) throws RuntimeException {
					return Boolean.TRUE.equals(c.getRequest().get("target"));
				}
			},
			new Fn<WebContext, Void, RuntimeException>() {
				public Void exec(WebContext c) throws RuntimeException {
					c.getRequest().put("status", "true");
					return null;
				}
			},
			new Fn<WebContext, Void, RuntimeException>() {
				public Void exec(WebContext c) throws RuntimeException {
					c.getRequest().put("status", "false");
					return null;
				}
			});
		
		final SimpleWebContext context = new SimpleWebContext();
		context.setRequest(new HashMap<String, Object>());
		context.setInput(new HashMap<String, List<Object>>());
		context.setSession(Utils.concurrentMap(
				new HashMap<String, Object>(), new Object()));
		
		context.setInput(Utils.map("transactionToken", Arrays.<Object>asList("token0")));
		context.getRequest().put("target", true);
		context.getSession().put("transactionToken", "token0");
		fn.exec(context);
		assertEquals("true", context.getRequest().get("status"));
		assertEquals(32, context.getSession().get("transactionToken").toString().length());

		context.getRequest().put("target", true);
		context.getSession().put("transactionToken", "token1");
		fn.exec(context);
		assertEquals("false", context.getRequest().get("status"));

		context.getRequest().put("target", true);
		context.getSession().remove("transactionToken");
		fn.exec(context);
		assertEquals("false", context.getRequest().get("status"));

		context.getRequest().put("target", false);
		context.getSession().put("transactionToken", "token0");
		fn.exec(context);
		assertEquals("true", context.getRequest().get("status"));

		context.getRequest().put("target", false);
		context.getSession().put("transactionToken", "token1");
		fn.exec(context);
		assertEquals("true", context.getRequest().get("status"));

		context.getRequest().put("target", false);
		context.getSession().remove("transactionToken");
		fn.exec(context);
		assertEquals("true", context.getRequest().get("status"));
	}
}
